from itsimodels.core.base_models import BaseModel, ChildModel
from itsimodels.core.fields import (
    BoolField,
    ListField,
    NumberField,
    StringField
)


class SearchMetric(ChildModel):
    key = StringField(required=True, alias='_key')

    title = StringField(required=True)

    aggregate_statop = StringField(default='avg')

    entity_statop = StringField(default='avg')

    fill_gaps = StringField(default='null_value')

    gap_custom_alert_value = StringField()

    gap_severity = StringField(default='unknown')

    gap_severity_color = StringField(default='')

    gap_severity_color_light = StringField(default='')

    gap_severity_value = StringField(default='-1')

    threshold_field = StringField(default='')

    unit = StringField(default='')


class KpiBaseSearch(BaseModel):
    key = StringField(required=True, alias='_key')

    base_search = StringField(required=True)

    title = StringField(required=True)

    alert_lag = NumberField(default=30)

    alert_period = StringField()

    description = StringField(default='')

    entity_filter_field = StringField(default='', alias='entity_id_fields')

    entity_split_field = StringField(default='', alias='entity_breakdown_id_fields')

    is_filter_entities_to_service = BoolField(default=False, alias='is_service_entity_filter')

    is_split_by_entity = BoolField(default=False, alias='is_entity_breakdown')

    metric_qualifier = StringField(default='')

    metrics = ListField(SearchMetric)

    search_alert_earliest = StringField(default='')



import inspect

from itsimodels.core.fields import (
    DictField,
    ListField,
    TypeField
)


class FieldDecoder(object):
    """Decodes raw data for a model"""

    def decode(self, field_name, field, data):
        """
        Decodes the data located at the given field

        :param field_name: the field name
        :param field: the model field
        :param data: the encoded object data
        :return: the model data for the given field
        """
        from itsimodels.core.base_models import BaseModel

        name = self.decode_field_name(field, field_name)

        field_value = data.get(name)
        if field_value is None:
            return field.decode(field_value)

        if isinstance(field, ListField) and inspect.isclass(field.subtype) and issubclass(field.subtype, BaseModel):
            return self.handle_list_of_models(field, field_value)

        if isinstance(field, DictField) and inspect.isclass(field.subtype) and issubclass(field.subtype, BaseModel):
            return self.handle_dict_of_models(field, field_value)

        if isinstance(field, TypeField) and inspect.isclass(field.type) and issubclass(field.type, BaseModel):
            model = field.type

            return model(data=field_value, field_decoder=self)

        return field.decode(field_value)

    def decode_field_name(self, field, field_name):
        return field_name

    def handle_list_of_models(self, field, lst):
        model_class = field.subtype
        models = []

        for model_data in lst:
            model = model_class(data=model_data, field_decoder=self)
            models.append(model)

        return models

    def handle_dict_of_models(self, field, dct):
        model_class = field.subtype
        models = {}

        for key, model_data in dct.items():
            model = model_class(data=model_data, field_decoder=self)
            models[key] = model

        return models



from itsimodels.core.base_models import BaseModel, ChildModel
from itsimodels.core.compat import string_types
from itsimodels.core.fields import (
    BoolField,
    DictField,
    ForeignKey,
    ForeignKeyList,
    ListField,
    NumberField,
    StringField,
    TypeField
)
from itsimodels.team import GLOBAL_TEAM_KEY


class ThresholdLevel(ChildModel):
    severity_color = StringField(alias='severityColor')

    severity_color_light = StringField(alias='severityColorLight')

    severity_label = StringField(alias='severityLabel')

    threshold_value = NumberField(alias='thresholdValue')


class KpiThresholds(ChildModel):
    base_severity_color = StringField(alias='baseSeverityColor')

    base_severity_color_light = StringField(alias='baseSeverityColorLight')

    base_severity_label = StringField(default='info', alias='baseSeverityLabel')

    base_severity_value = NumberField(alias='baseSeverityValue')

    gauge_max = NumberField(default=100, alias='gaugeMax')

    gauge_min = NumberField(default=0, alias='gaugeMin')

    is_max_static = BoolField(default=False, alias='isMaxStatic')

    is_min_static = BoolField(default=False, alias='isMinStatic')

    metric_field = StringField(alias='metricField')

    render_boundary_max = NumberField(alias='renderBoundaryMax')

    render_boundary_min = NumberField(alias='renderBoundaryMin')

    threshold_levels = ListField(ThresholdLevel, alias='thresholdLevels')


class KpiTimePolicy(ChildModel):
    title = StringField(required=True)

    aggregate_thresholds = TypeField(KpiThresholds)

    entity_thresholds = TypeField(KpiThresholds)

    policy_type = StringField(default='static')

    time_blocks = ListField(object)


class KpiTimePolicies(ChildModel):
    policies = DictField(KpiTimePolicy)


class MetricSearchSpec(ChildModel):
     metric_index = StringField(default='')

     metric_name = StringField(default='')


class Kpi(ChildModel):
    key = StringField(required=True, alias='_key')

    title = StringField(required=True)

    aggregate_statop = StringField(default='avg')

    adaptive_thresholds_is_enabled = BoolField(default=False)

    aggregate_thresholds = TypeField(KpiThresholds)

    aggregate_thresholds_custom_alert_rules = ListField()

    alert_lag = StringField(default='30')

    alert_on = StringField()

    alert_period = StringField()

    anomaly_detection_alerting_enabled = BoolField(default=False)

    anomaly_detection_is_enabled = BoolField(default=False)

    anomaly_detection_sensitivity = NumberField()

    anomaly_detection_training_window = StringField()

    aggregate_threshold_alert_enabled = BoolField(default=False)

    backfill_earliest_time = StringField(default='-7d')

    backfill_enabled = BoolField(default=False)

    base_search = StringField()

    base_search_id = ForeignKey('itsimodels.kpi_base_search.KpiBaseSearch')

    base_search_metric = ForeignKey('itsimodels.kpi_base_search.SearchMetric')

    cohesive_ad = DictField()

    cohesive_anomaly_detection_is_enabled = BoolField(default=False)

    description = StringField(default='')

    enabled = BoolField(default=False)

    entity_filter_field = StringField(default='', alias='entity_id_fields')

    entity_split_field = StringField(default='', alias='entity_breakdown_id_fields')

    entity_statop = StringField(default='avg')

    entity_thresholds = TypeField(KpiThresholds)

    fill_gaps = StringField(default='null_value')

    gap_custom_alert_value = NumberField()

    gap_severity = StringField(default='unknown')

    gap_severity_color = StringField(default='')

    gap_severity_color_light = StringField(default='')

    gap_severity_value = StringField(default='-1')

    kpi_template_kpi_id = ForeignKey('itsimodels.kpi_threshold_template.KpiThresholdTemplate')

    kpi_threshold_template_id = StringField()

    is_filter_entities_to_service = BoolField(default=False, alias='is_service_entity_filter')

    is_split_by_entity = BoolField(default=False, alias='is_entity_breakdown')

    metric_search_spec = TypeField(MetricSearchSpec, alias='metric')

    metric_qualifier = StringField()

    search = StringField()

    search_alert = StringField()

    search_alert_earliest = StringField()

    search_alert_entities = StringField()

    search_buckets = StringField()

    search_occurrences = NumberField()

    search_type = StringField(default='adhoc')

    threshold_field = StringField(default='')

    time_policies = TypeField(KpiTimePolicies, alias='time_variate_thresholds_specification')

    trending_ad = DictField()

    type = StringField(default='kpis_primary')

    tz_offset = StringField()

    unit = StringField(default='')

    use_time_policies = BoolField(default=False, alias='time_variate_thresholds')

    urgency = NumberField()


class EntityRuleItem(ChildModel):
    field = StringField()

    field_type = StringField()

    rule_type = StringField()

    value = StringField()


class EntityRule(ChildModel):
    rule_condition = StringField()

    rule_items = ListField(EntityRuleItem)


class ServiceDependency(ChildModel):
    kpis_depending_on = ForeignKeyList('itsimodels.service.Kpi')

    service_id = ForeignKey('itsimodels.service.Service', alias='serviceid')


class ServiceTags(ChildModel):
    tags = ListField(string_types)

    template_tags = ListField(string_types)


class Service(BaseModel):
    key = StringField(required=True, alias='_key')

    title = StringField(required=True)

    description = StringField(default='')

    enabled = BoolField(default=False)

    entity_rules = ListField(EntityRule)

    kpis = ListField(Kpi)

    services_depends_on = ListField(ServiceDependency)

    services_depending_on_me = ListField(ServiceDependency)

    service_template_id = ForeignKey('itsimodels.service_template.ServiceTemplate', alias='base_service_template_id')

    service_tags = TypeField(ServiceTags)

    team_id = ForeignKey('itsimodels.team.Team', default=GLOBAL_TEAM_KEY, alias='sec_grp')



from itsimodels.core.base_models import BaseModel, ChildModel
from itsimodels.core.fields import (
    BoolField,
    NumberField,
    StringField,
    TypeField
)


class TileSettings(ChildModel):
    is_filter_enabled = BoolField(default=False, alias='isFilterEnabled')

    num_tiles = NumberField(default=50, alias='numTiles')


class ServiceAnalyzer(BaseModel):
    key = StringField(required=True, alias='_key')

    title = StringField(required=True)

    description = StringField(default='')

    earliest_time = StringField()

    kpi_filter_string = StringField(default='', alias='kpiFilterString')

    kpi_tiles_settings = TypeField(TileSettings, alias='kpiTilesSettings')

    latest_time = StringField()

    search_type = StringField(default='aggregate', alias='searchType')

    service_filter_string = StringField(default='', alias='serviceFilterString')

    service_tiles_settings = TypeField(TileSettings, alias='serviceTilesSettings')

    tile_size = StringField(
        choices=('small', 'medium', 'large'),
        default='large',
        alias='tileSize')

    view = StringField(default='view')

    view_type = StringField(
        choices=('tile', 'topology'),
        default='tile',
        alias='viewType')



import collections
import copy

from itsimodels import __version__
from itsimodels.core.compat import string_types
from itsimodels.core.field_decode import FieldDecoder
from itsimodels.core.fields import (
    BaseField,
    DictField,
    FieldValidationError,
    ListField,
    StringField,
    TypeField
)
from itsimodels.core.migrators import Migrator


KEY_FIELD_NAME = 'key'


class ValidationError(Exception):
    """Raised when a data or validation error occurs on the model"""


class BaseModel(object):
    """Represents an ITSI knowledge object"""

    # Helper object to handle data migrations between different model versions
    migrator = Migrator()

    # The model version
    version = StringField(default=__version__, required=True)

    @classmethod
    def model_fields(cls):
        """
        Returns a mapping dict of field name to field types

        :return: a dict from str to an instance of BaseField
        :rtype: dict
        """
        fields = {}

        class_items = list(BaseModel.__dict__.items()) + list(cls.__dict__.items())
        for name, value in class_items:
            if not isinstance(value, BaseField):
                continue

            fields[name] = value

        return fields

    def __init__(self, data=None, auto_validate=True, auto_migrate=True, field_decoder=None):
        """
        :param dict data: The model data
        :param bool auto_validate: Whether to automatically call validate on the model
        :param bool auto_migrate: Whether to automatically migrate the model's data to the latest model version
        :param object field_decoder: Optional object to help parse a field value
        """
        if data is None:
            data = {}

        if data.get('version', None) is None:
            data['version'] = self.version.default

        if auto_migrate:
            data = self.migrator.migrate(data)

        self._fields = self._populate(data, field_decoder=field_decoder)

        if auto_validate:
            self.validate()

    def get_key(self):
        return getattr(self, KEY_FIELD_NAME, None)

    def validate(self):
        """
        Validates the current model based on its schema of fields.
        """
        for name, field in self._fields.items():
            value = getattr(self, name)
            try:
                field.validate(value)
            except FieldValidationError as exc:
                raise ValidationError('"{}" {} for {}'.format(name, exc, self))

    @property
    def fields(self):
        return self._fields

    def to_dict(self, use_alias=False):
        """
        Returns a dict object that represents the given model.

        :param use_alias: whether to use the alias name instead
        :type use_alias: bool

        :return: a dict containing the model's data
        :rtype: dict
        """
        obj = {}

        for name, field in self.fields.items():
            field_value = getattr(self, name)

            if (field_value and isinstance(field, TypeField) and
                    field.type and issubclass(field.type, BaseModel)):
                copied_value = field_value.to_dict(use_alias=use_alias)

            elif (field_value and isinstance(field, ListField) and
                  field.subtype and issubclass(field.subtype, BaseModel)):
                values = []

                for index, value in enumerate(field_value):
                    values.append(value.to_dict(use_alias=use_alias))

                copied_value = values

            elif (field_value and isinstance(field, DictField) and
                  field.subtype and issubclass(field.subtype, BaseModel)):
                values = {}

                for key, value in field_value.items():
                    values[key] = value.to_dict(use_alias=use_alias)

                copied_value = values
            else:
                copied_value = copy.deepcopy(field_value)

            field_name = field.alias if use_alias and field.alias else name

            obj[field_name] = copied_value

        return obj

    def _populate(self, data, field_decoder=None):
        """
        Populates the model with values from data for each field

        :param dict data: a dict from field name to data value
        :param object field_decoder: Optional object to help parse a field value
        """
        fields = self.model_fields()

        return self._populate_with_fields(data, fields, field_decoder=field_decoder)

    def _populate_with_fields(self, data, fields, field_decoder=None):
        """
        Populates the model with values from data for each field in the given set of fields

        :param dict data: a dict from field name to data value
        :param object field_decoder: Optional object to help parse a field value
        """
        if field_decoder is None:
            field_decoder = FieldDecoder()

        for field_name, field in fields.items():
            default = field.default

            value = field_decoder.decode(field_name, field, data)

            if value is None and default is not None:
                if isinstance(default, collections.Callable):
                    value = default()
                else:
                    value = copy.deepcopy(default)

            setattr(self, field_name, value)

        return fields


class ChildModel(BaseModel):
    """Represents an ITSI knowledge object that's nested in another model"""

    def __init__(self, *args, **kwargs):
        super(ChildModel, self).__init__(*args, **kwargs)

        if hasattr(self, 'version'):
            delattr(self, 'version')
            self._fields.pop('version', None)


class DynamicModel(BaseModel):
    """Represents an object with a dynamic field set"""

    def to_dict(self, *args, **kwargs):
        """
        Returns a dict obj that represents the current model

        :return: a dict
        :rtype: dict
        """
        data = {}

        for field in dir(self):
            if field.startswith('__') or field.startswith('_') or field in ['fields']:
                continue

            value = getattr(self, field)
            if isinstance(value, (string_types, int, float, dict, list)):
                data[field] = value

        return data

    def _populate(self, data, field_decoder=None):
        """
        Populates the model with values from data for each field

        :param dict data: a dict from field name to data value
        :param object field_decoder: Optional object to help parse a field value
        """
        fields = {}

        key = self.get_key()
        if key is not None:
            fields[KEY_FIELD_NAME] = key

        for field_name, value in data.items():
            if isinstance(value, list):
                field = ListField()
            elif isinstance(value, dict):
                field = DictField()
            else:
                field = StringField()

            fields[field_name] = field

        fields.update(self.model_fields())

        return self._populate_with_fields(data, fields, field_decoder=field_decoder)


class ConfModel(DynamicModel):
    """Represents a conf object"""


class ImageModel(BaseModel):
    """Represents an image object"""



from itsimodels.core.base_models import (
    BaseModel,
    ChildModel
)
from itsimodels.core.compat import string_types
from itsimodels.core.fields import (
    BoolField,
    DictField,
    ListField,
    StringField,
    TypeField
)


class DataDrilldown(ChildModel):
    title = StringField(required=True)

    type = StringField(required=True)

    entity_field_filter = DictField(object)

    static_filter = DictField(object)


class DashboardDrilldownParams(ChildModel):
    alias_param_map = ListField(object)

    static_params = DictField(object)


class DashboardDrilldown(ChildModel):
    title = StringField(required=True)

    id = StringField(required=True)

    base_url = StringField(default='')

    is_splunk_dashboard = BoolField(default=False)

    params = TypeField(DashboardDrilldownParams)


class VitalMetric(ChildModel):
    metric_name = StringField(required=True)

    search = StringField(required=True)

    is_key = BoolField(default=False)

    matching_entity_fields = ListField(string_types)

    split_by_fields = ListField(string_types)

    unit = StringField(default='')


class EntityType(BaseModel):
    key = StringField(required=True, alias='_key')

    title = StringField(required=True)

    description = StringField(default='')

    vital_metrics = ListField(VitalMetric)

    data_drilldowns = ListField(DataDrilldown)

    dashboard_drilldowns = ListField(DashboardDrilldown)



from itsimodels.core.base_models import BaseModel, ChildModel
from itsimodels.core.compat import string_types
from itsimodels.core.fields import (
    BoolField,
    DictField,
    ForeignKey,
    ForeignKeyList,
    ListField,
    NumberField,
    StringField,
    TypeField
)


KPI_SUMMARY_RE = r'`get_full_itsi_summary_kpi\((.+)\)`'


KV_STORE_KEY_RE = r'splunk-enterprise-kvstore://(.+)'


class ValuedModel(ChildModel):

    def to_dict(self, *args, **kwargs):
        obj = super(ValuedModel, self).to_dict(*args, **kwargs)

        # Remove any null values to conform with the UI visualization library
        data = {key: value for key, value in obj.items() if value}

        return data


class DataSourceMeta(ChildModel):
    kpi_id = ForeignKey('itsimodels.service.Kpi', alias='kpiID')

    service_id = ForeignKey('itsimodels.service.Service', alias='serviceID')


class DataSourceOptions(ValuedModel):
    data = DictField()

    query = ForeignKey('itsimodels.service.Kpi', key_regex=KPI_SUMMARY_RE)


class DataSource(ValuedModel):
    meta = TypeField(DataSourceMeta)

    name = StringField()

    options = TypeField(DataSourceOptions)

    primary = StringField()

    type = StringField()


class DefinitionInput(ChildModel):
    options = DictField()

    type = StringField()


class LayoutImage(ChildModel):
    size_type = StringField(alias='sizeType')

    src = ForeignKey('itsimodels.glass_table_image.GlassTableImage', key_regex=KV_STORE_KEY_RE)

    x = NumberField()

    y = NumberField()


class LayoutOptions(ChildModel):
    background_color = StringField(alias='backgroundColor')

    background_image = TypeField(LayoutImage, alias='backgroundImage')

    height = NumberField()

    width = NumberField()


class StructureItem(ChildModel):
    item = StringField()

    position = DictField()

    type = StringField()


class Layout(ChildModel):
    global_inputs = ListField(string_types, alias='globalInputs')

    options = TypeField(LayoutOptions)

    structure = ListField(StructureItem)

    type = StringField()


class VisualizationOptions(ValuedModel):
    background_color = StringField(alias='backgroundColor')

    color = StringField()

    content = StringField()

    fill = StringField()

    font_size = NumberField(alias='fontSize')

    icon = ForeignKey('itsimodels.glass_table_icon.GlassTableIcon', key_regex=KV_STORE_KEY_RE)

    number_precision = NumberField(alias='numberPrecision')

    preserve_aspect_ratio = BoolField(alias='preserveAspectRatio')

    rx = NumberField()

    ry = NumberField()

    show_value = BoolField(alias='showValue')

    src = ForeignKey('itsimodels.glass_table_image.GlassTableImage', key_regex=KV_STORE_KEY_RE)

    stroke = StringField()

    stroke_width = NumberField(alias='strokeWidth')


class Visualization(ValuedModel):
    data_sources = DictField(alias='dataSources')

    options = TypeField(VisualizationOptions)

    primary = StringField()

    type = StringField()


class Definition(ChildModel):
    data_sources = DictField(DataSource, alias='dataSources')

    inputs = DictField(DefinitionInput)

    layout = TypeField(Layout)

    visualizations = DictField(Visualization)


class GlassTable(BaseModel):
    key = StringField(required=True, alias='_key')

    title = StringField(required=True)

    definition = TypeField(Definition)

    description = StringField(default='')

    gt_version = StringField(default='beta')

    latest = StringField(default='now')

    latest_label = StringField(default='Now')

    template_selected_service_id = ForeignKey('itsimodels.service.Service', alias='templateSelectedServiceId')

    template_swappable_service_ids = ForeignKeyList('itsimodels.service.Service', alias='templateSwappableServiceIds')



import json
import os

from itsimodels.core.field_decode import BackupFieldDecoder
from itsimodels.correlation_search import *
from itsimodels.deep_dive import *
from itsimodels.entity_type import *
from itsimodels.glass_table import *
from itsimodels.glass_table_icon import *
from itsimodels.glass_table_image import *
from itsimodels.kpi_base_search import *
from itsimodels.kpi_threshold_template import *
from itsimodels.neap import *
from itsimodels.service import *
from itsimodels.service_analyzer import *
from itsimodels.service_template import *
from itsimodels.team import *


HERE = os.path.dirname(os.path.realpath(__file__))


def test_service_model():
	fixture = os.path.join(HERE, 'fixtures', 'backup', 'services.json')

	with open(fixture) as fobj:
		raw_objects = json.loads(fobj.read())

	objects = []
	for raw_data in raw_objects:
		decoded = Service(raw_data, field_decoder=BackupFieldDecoder())
		objects.append(decoded)

	assert len(objects) == 36

	for model in objects:
		raw_data = model.to_dict()

		service = Service(raw_data)

		backup = model.to_dict(use_alias=True)
